//#target "Illustrator"
//#targetengine "ObjLoaderSession"
//$.locale = "en";  // アンコメントすると英語UI

// OBJ形式のファイルを読み込むスクリプトです

// 基本的な使い方
// 1. "OBJ読み込み" ボタンで obj ファイルを選択する
//   （読み込んだ形状は書類の原点(0,0)を中心として作成されます）
// 2. 向きを変えたいときはダイアログ上部のパネル上でドラッグする
// 3. ダイアログを閉じる

// 補足
// ・ドラッグ操作においては、ドラッグの開始点が原点（視点の中心）になり
//  ます。（パネルの中心が原点ではありません。）
//
// ・マテリアルはディフューズ色のみ反映されます。データを読み込むごと
//   に、スウォッチにRGBのスポットカラーがマテリアル数分だけ追加されます。

// 進んだ使い方
// ・"光源位置設定" にチェックをつけると、パネル上のドラッグで光源の位置
//   を変更できます。チェックはドラッグ終了時に外れます。
//
// ・"オプション" ボタンを押して設定を変更できます（詳細後述）
//   オプションのデフォルト値を変更するにはスクリプトの冒頭部分を編集し
//   てください。
//
// ・"影をつける" ボタンを押して影をつけられます。(影がついている状態で
//   回転はできません。改めて回転させるには "影を削除" ボタンを押して影
//   を削除してください。）
//   影は面ごとに作成しています。影が重なった部分は、追ってパスファインダ
//   で合体するなどしてください。

// オプションの補足
// ・ドラッグ中は簡易的なZ座標補正を行う：「Z座標補正」は面の前後の並べ替え処
//   理で、データ量が多いと時間がかかります。ドラッグ中は簡易的な並べ替えを行
//   うよう設定できます。簡易処理を行っている際は前後関係が正しく表示されない
//   ことがあります。必要に応じてメインダイアログの「Z座標補正」ボタンを押し
//   てください。影をつける時にはここでの設定にかかわらず簡易的でない並べ替え
//   処理が行われます。
//
// ・終了時に裏面を残す：既定では視線方向と反対側を向いた面は終了時に削
//   除されます。この設定をチェックすると、これらの面はグループ化され、
//   読み込みで作成したパスの最背面に置かれます。
//
// ・チント下限：面の明るさはスポットカラーのチントで設定しています。下
//   限は最も明るい部分の濃度です。
//
// ・大きいデータと見なす閾値：設定した面数以上のデータを読み込んだ場
//   合、警告などを表示します。
//
// ・点光源：光の減衰はありません。
//
// ・スクリプトの冒頭部に、パネルで設定できないオプションがいくつかあります。
//   たとえば既定ではx座標またはy座標が原点から1000ポイント以上離れた点をもつ
//   パスは描画されません。用途に合わない場合は _MAX_COORD_VALUE の値を変更し
//   てください。

// Copyright(c) 2017 Hiroyuki Sato
// http://shspage.hatenablog.com/
// https://github.com/shspage
// MITライセンスのフリーウェアです。
// 詳細は同梱のLICENSE.txtをご覧ください。

// ver.1.0.1 ... 未使用変数削除、コメント修正
// ver.1.0.2 ... UI(メッセージ)を一部変更
// ver.1.0.4 ... 一部処理を簡略化
// ver.1.0.5 ... mouseoutのeventListenerを追加
// ver.1.1.0 ... mouseUp時のz座標補正を廃止。z座標補正ボタンを追加
// ver.1.1.1 ... 冒頭の説明文を現バージョンに合わせて変更
// ver.1.1.2 ... 閉じるボタンを追加

var ObjLoader;
//if(!ObjLoader)
(function(){
    var _PANEL_SIZE = 200;          // パネル（ドラッグ面）の辺の長さ(pixel)
    var _MIN_DRAG_DIST = 16;        // 前回描画位置からの距離の２乗

    var _TINT_MAX = 100;            // 面色のチントの最大値
    var _MAX_COORD_VALUE = 1000;    // 出力座標値x,yの絶対値の最大値(point)
    
    var _HPI = Math.PI / 2;         // 90度
    var _EPSILON = 1e-12;           // 数値が等しいと見なす誤差の許容値
    
    var _DEBUG = true;

    var _sources = [];              // 回転の対象となるオブジェクトの情報を保持(ItemInfo[])
    var _center;                    // 回転の（グローバルな）中心(Point3)。
                                    // 現在は(0,0,0)。これ以外にする場合は何箇所か処理の変更が必要。
    var _radius = _PANEL_SIZE / 2;  // ドラッグの距離を角度に換算する際に使用
    var _light;                     // 光源の位置(Point3)
    var _last_quaternion;           // ドラッグ中の最後の回転
    var _base_quaternion;           // ドラッグ開始前に適用されている回転
    var _vz;                        // new Point3(0,0,1);
    var _win;                       // メインダイアログ
    var _is_large_data;             // large_data_threshold 以上のデータの場合 true

    // OBJファイルのデータ
    var _obj_vs;                    // 頂点データ Point3[]
    var _obj_tmp_vs;                // 回転後の座標 Point3[]
    var _obj_view_vs;               // 投影後の座標 Point[]
    var _obj_groups;                // ObjGroup[]

    // 影
    var _shadow_color;  // Color. 影の色
    var _shadow_group_ground;  // GroupItem. 地面に落ちた影の形状を入れたgroup

    var _spec = {
        vmin : undefined,           // 全体のx,y,z座標の最小値を保持
        vmax : undefined            // 全体のx,y,z座標の最大値を保持
    };

    // オプション
    var _opt = {
        projection: 0,              // 投影法（0:平行投影、1:透視投影）
        focal_length: 500,          // 焦点距離（透視投影で使用）
        scale_type : 0,             // 拡大方法（0:自動、1:指定)
        scale_factor : 10,          // 読み込んだ座標値に掛ける倍率。
        size : 50,                  // scaleが自動のときのサイズ(mm)
        tint_min: 10,               // 面色のチントの最小値
        save_hidden_surfaces: false,  // ダイアログ終了時に陰面を残す
        light_type : 0,             // 0:平行光源, 1:点光源
        light_distance : 100,       // 点光源の距離
        use_simple_z_sort_while_dragging : false,  // ドラッグ中は簡易的なZ座標補正を行う
        shadow_gray_value : 50,     // 影の濃さ(GrayColor.grayの値)
        large_data_threshold : 100, // 大容量データ用の処理を行う閾値（面の数）
        backlight_ratio : 0.2,      // 光源の逆方向から当てる光の割合
        // 以下はダイアログにない項目
        cast_shadow_on_hidden_faces : false,  // 光源と逆向きの面にも影を作成する
        warning_on_large_data_load : true,    // 大容量データの読み込み時に警告する
        cast_shadow_ground_angle_threshold : 5,  // 光源の水平面からの角度がこれ以下の場合、地面への影は描画しない(度数法)
    };

    var _flags = {                  // 内部的に使うフラグ
        drawn : false,              // スクリプトによって描画を更新したかどうか（undoする条件）
        drawn_light : false,        // 光源が変更されたかどうか（undoする条件）
        mode : 0,                   // 0:通常描画, 1:リセット, 2:陰面も描画する
    };

    // TODO: ダイアログを閉じたあとのundoで実行前に戻るようにする
    // TODO: CMYKColorの対応？
    // TODO: 読み込み直後にドラッグするとqmul？がエラーになる件

    // includefile objLoader_loadObj.jsxinc begin
    // OBJの面の情報を保持する
    var ObjFace = function(){
        this.idxs = [];  // int[]. 頂点index
    }
    
    // 
    var ObjGroup = function(){
        this.faces = [];  // ObjFace[]
        this.material;  // マテリアル名
    }
    
    // マテリアルの情報を保持する。
    // いまのところディフューズ色しか反映しない。
    var ObjMaterial = function(){
        // Ka アンビエント色 (0-1, RGB)
        // Kd ディフューズ色
        // Ks スペキュラ色
        //    Ns スペキュラ指数(0-1000)
        // d(またはTr) 透過(Trは反転)
        // illum
        // 0. 色 有効 かつ アンビエント 無効
        // 1. 色 有効 かつ アンビエント 有効
        // 2. ハイライト 有効
        // 3. 反射 有効 かつ レイトレース 有効
        // 4. 透過: グラス 有効, 反射: レイトレース 有効
        // 5. 反射: フレネル 有効 かつ レイトレース 有効
        // 6. 透過: 反射 有効, 反射: フレネル 無効 かつ レイトレース 有効
        // 7. 透過: 反射 有効, 反射: フレネル 有効 かつ レイトレース 有効
        // 8. 反射 有効 かつ レイトレース 無効
        // 9. 透過: ガラス 有効, 反射: レイトレース 無効
        // 10. 不可視サーフィスに影を投影
        this.Kd;  // RGB, またはCMYK Color
    }
    ObjMaterial.prototype = {
        // material _name: マテリアル名。スポットカラーの名称にするが、
        //     同じドキュメント内で重複した名前を使えないため、
        //     重複する場合は末尾に連番を付加する。
        // r: ["Kd", r, g, b], [0]は不使用、rgbは(0.0-1.0)
        // current_spots: 現在の書類のスポットカラー名をキーにした辞書。getCurrentSpots()で生成されるもの
        setColor : function(material_name, r, current_spots){
            var c = new RGBColor();
            c.red = tof(r[1]) * 255;
            c.green = tof(r[2]) * 255;
            c.blue = tof(r[3]) * 255;
            
            var spot = activeDocument.spots.add();
            spot.color = c;
            spot.name = getSpotName(material_name, current_spots);
            
            var sp = new SpotColor();
            sp.spot = spot;
            sp.colorType = ColorModel.PROCESS;
    
            this.Kd = sp;
        },
        toString : function(){
            return "Material";
        }
    }
    
    // 既定のマテリアル（スポットカラーを生成して返す。
    // current_spots: 現在の書類のスポットカラー名をキーにした辞書。undefinedの場合新たに作成
    function ObjGetDefaultMaterial(current_spots){
        if(!current_spots) current_spots = getCurrentSpots();
        var mat = new ObjMaterial();
        var spot_name = getSpotName("Material_default", current_spots);
        mat.setColor("Material_default", [0,0.5,0.5,0.5], current_spots);
        return mat;
    }
    
    // 現在の書類中で重複しないスポットカラー名を生成して返す。
    // material_name: マテリアル名
    // current_spots: 現在の書類のスポットカラー名をキーにした辞書。getCurrentSpots()で生成されるもの
    function getSpotName(material_name, current_spots){
        if(current_spots){
            if(material_name in current_spots){
                var count = 1;
                while((material_name + "." + count) in current_spots){
                    count++;
                }
                material_name += "." + count;
            }
        }
        current_spots[material_name] = 1;
        return material_name;
    }
    
    // OBJファイルを読み取る。
    function ObjLoad(){
        var splitLine = function(s){
            return s.replace("\r","").split(" ");
        }
    
        var data;  // ファイルの内容
        try{
            // ファイル選択ダイアログ
            var f = File.openDialog(localize(resources.message.invalid_file_selection));
            if(f == null) return;
            
            if(!f.open("r")){
                alert(localize(resources.message.fail_to_open_obj_to_load));
                return;
            }
            
            data = f.read();
            f.close();
            
            showPercent(5);
        } catch(e){
            alert(localize(resources.message.error_on_loading_obj) + e.message);
            return;
        }
    
        var lines = data.split("\n");
        
        _obj_vs = [];  // 頂点座標の Point3[]
        var group;       // ObjGroup インスタンス
        var groups = [];  // ObjGroup[]
        var faces_count = 0;
    
        var material_files = [];  // objファイル中に出現したマテリアルファイル名を保持
        
        for(var i = 0, iEnd = lines.length; i < iEnd; i++){
            var line = lines[i];
            if(line.length < 2) continue;
    
            if(line.indexOf("mtllib") == 0){
                material_files.push(splitLine(line)[1]);
            } else if(line.indexOf("usemtl") == 0){
                if(group) group.material = splitLine(line)[1];
            } else {
            
                var key = line.substring(0, 2);
                
                if(key == "g " || key == "o "){  // group の開始
                    if(group && group.faces.length > 0) groups.push(group);
                    group = new ObjGroup();
                } else if(key == "v "){  // 頂点座標 vertex
                    var r = splitLine(line);
                    _obj_vs.push(new Point3(tof(r[1]), tof(r[2]), tof(r[3])));
                } else if(key == "f "){  // 面 face
                    if(group){
                        var r = splitLine(line);
                        var face = new ObjFace();
                        for(var j = 1; j < r.length; j++){
                            var r1 = r[j].split("/");
                            var idx = toi(r1[0]) - 1;
                            face.idxs.push(idx);
                        }
                        group.faces.push(face);
                        faces_count++;
                    }
                }
            }
        }
        if(group && group.faces.length > 0) groups.push(group);
    
        _is_large_data = faces_count > _opt.large_data_threshold;
        if(_opt.warning_on_large_data_load
           && _is_large_data){
            var msg = localize(resources.message.warning_on_large_data_load)
              + "\r(" + faces_count.toString() + " faces)";
            
            if(_opt.use_simple_z_sort_while_dragging == false){
                msg += "\r" + localize(resources.message.warning_on_large_data_load_plus);
            }
            
            if(confirm(msg) == false) return;
    
            if(_opt.use_simple_z_sort_while_dragging == false){
                _opt.use_simple_z_sort_while_dragging = true;
            }
        }
    
        // マテリアルファイルの読み取り
        if(groups.length > 0 && material_files.length > 0){
            var current_spots = getCurrentSpots();  // キー：スポットカラー名, 値：1 (不使用)
            
            var materials = {};  // キー：マテリアル名、値：ObjMaterial インスタンス
            
            var r_dir_path = f.absoluteURI.split("/");
            r_dir_path.pop();
            var dir_path = r_dir_path.join("/") + "/";
            
            for(var i = 0; i < material_files.length; i++){
                var mat_file = material_files[i];
    
                // パスを取得
                var fm = new File(dir_path + mat_file);
                // ファイルの存在確認
                try {
                    if(!fm.open("r")){
                        alert(localize(resources.message.fail_to_open_mtl_to_load));
                        return;
                    }
                    
                    data = fm.read();
                    fm.close();
                } catch(e){
                    alert(localize(resources.message.error_on_loading_obj) + e.message);
                    return;
                }
    
                var material_name;
                var lines = data.split("\n");
                
                for(var li = 0, liEnd = lines.length; li < liEnd; li++){
                    var line = lines[li];
                    if(line.length < 2) continue;
    
                    if(line.indexOf("newmtl") == 0){
                        material_name = splitLine(line)[1];
                        materials[material_name] = new ObjMaterial();
                    } else {
                        
                        var key = line.substring(0, 2);
    
                        if(key == "Kd"){
                            var r = splitLine(line);
                            materials[material_name].setColor(material_name, r, current_spots);
                        }
                    }
                }
            }
    
            // 適用
            var default_material;
            
            for(var i = 0, iEnd = groups.length; i < iEnd; i++){
                if(groups[i].material && groups[i].material in materials){
                    groups[i].material = materials[ groups[i].material ];
                } else {
                    if(!default_material) defaul_material = ObjGetDefaultMaterial(current_spots);
                    groups[i].material = defaul_material;
                }
            }
        }
    
        showPercent(10);
        return groups;
    }
    
    // 現在の書類中のスポットカラー名の辞書を作成して返す
    function getCurrentSpots(){
        var current_spots = {};
        var spots = activeDocument.spots;
        if(spots){
            for(var i = 0, iEnd = spots.length; i < iEnd; i++){
                current_spots[spots[i].name] = 1;
            }
        }
        return current_spots;
    }
    
    // OBJファイルの内容に基づきartboard上に描画する
    function ObjRender(){
        var groups = ObjLoad();
        if(groups == null || !groups || groups.length < 1){
            alert(localize(resources.message.nothing_loaded));
            return;
        }
    
        getCenterOfSources(groups);
        _center = new Point3(0,0,0);
    
        var progress_start = 15;
        showPercent(progress_start);
        var progress_remain = 100 - progress_start;
        
        _sources = [];
    
        var pitems = activeDocument.activeLayer.pathItems;
        var faces_count = 0;
        var faces_total = countFaces(groups);
        var default_material;
    
        var gitem = activeDocument.activeLayer.groupItems.add();
    
        for(var i = 0, iend = groups.length; i < iend; i++){
            var group = groups[i];
            
            for(var fi = 0, fiend = group.faces.length; fi < fiend; fi++){
                var face = group.faces[fi];
                
                var p = pitems.add();
                p.closed = true;
                p.stroked = false;
                if(group.material){
                    p.fillColor = group.material.Kd;
                } else {
                    if(!default_material){
                        default_material = ObjGetDefaultMaterial();
                    }
                    p.fillColor = default_material.Kd;
                }
                p.move(gitem, ElementPlacement.PLACEATBEGINNING);
    
                for(var vi = 0, viend = face.idxs.length; vi < viend; vi++){
                    var pp = p.pathPoints.add();
                    var idx = face.idxs[vi];
                    var ar = _obj_vs[idx].toArray();
                    fixCoord(ar);
                    pp.anchor = ar;
                    pp.rightDirection = ar;
                    pp.leftDirection = ar;
                }
                
                //p.selected = true;
                var info = new ItemInfo(p, face.idxs, group);
                _sources.push(info);
    
                faces_count++;
                showPercent((faces_count / faces_total) * progress_remain + progress_start);
            }
        }
    
        return groups;
    }
    
    // 面の総数を返す
    function countFaces(groups){
        var c = 0;
        for(var i = 0, iend = groups.length; i < iend; i++){
            c += groups[i].faces.length;
        }
        return c;
    }
    
    // 全体の中心を求める
    function getCenterOfSources(groups){
        var vmin = _obj_vs[0].clone();
        var vmax = _obj_vs[0].clone();
        
        for(var vi = 0, viEnd = _obj_vs.length; vi < viEnd; vi++){
            var v = _obj_vs[vi];
            
            if(v.x < vmin.x){
                vmin.x = v.x;
            } else if(v.x > vmax.x){
                vmax.x = v.x;
            }
            if(v.y < vmin.y){
                vmin.y = v.y;
            } else if(v.y > vmax.y){
                vmax.y = v.y;
            }
            if(v.z < vmin.z){
                vmin.z = v.z;
            } else if(v.z > vmax.z){
                vmax.z = v.z;
            }
        }
    
        var center = new Point3((vmin.x + vmax.x) / 2,
                                (vmin.y + vmax.y) / 2,
                                (vmin.z + vmax.z) / 2);
        
        ObjScale(groups, vmin, vmax, center);
        
        return center;
    }
    
    // 設定に基づき全体を拡大・縮小する
    function ObjScale(groups, vmin, vmax, center){
        var m = _opt.scale_factor;
        if(_opt.scale_type == 0){
            var x_size = vmax.x - vmin.x;
            var y_size = vmax.y - vmin.y;
            var z_size = vmax.z - vmin.z;
            var s = Math.max(z_size, Math.max(x_size, y_size));
            m = mm2pt(_opt.size) / s;
        }
        
        for(var vi = 0, viEnd = _obj_vs.length; vi < viEnd; vi++){
            _obj_vs[vi] = _obj_vs[vi].sub(center).mul(m);
        }
    
        _spec.vmin = vmin.sub(center).mul(m);
        _spec.vmax = vmax.sub(center).mul(m);
    }
    // objLoader_loadObj.jsxinc end

    /* ------------------------------------
     * Point：２次元の座標値を扱う
     * ------------------------------------ */
    var Point = function(x, y){
        this.x = x;
        this.y = y;
    }
    Point.prototype = {
        // v : Pointまたは .x,.y を持つもの
        add : function(v){
            return new Point(this.x + v.x, this.y + v.y);
        },
        sub : function(v){
            return new Point(this.x - v.x, this.y - v.y);
        },
        mul : function(m){
            return new Point(this.x * m, this.y * m);
        },
        cross : function(v){
            return this.x * v.y - v.x * this.y;
        },
        // 配列を返す。anchorなどへの設定用
        toArray : function(){
            return [this.x, this.y];
        },
        toString: function(){
            return "[" + this.x + "," + this.y + "]";
        }
    }
    
    /* ------------------------------------
     * Point3：３次元の座標値を扱う
     * ------------------------------------ */
    var Point3 = function(x, y, z){
        this.x = x;
        this.y = y;
        this.z = z;
    }
    Point3.prototype = {
        clone : function(){
            return new Point3(this.x, this.y, this.z);
        },
        fromArray : function(r){  // anchorなどから作成
            return new Point3(r[0], r[1], 0);
        },
        add : function(v){
            return new Point3(this.x + v.x, this.y + v.y, this.z + v.z);
        },
        sub : function(v){
            return new Point3(this.x - v.x, this.y - v.y, this.z - v.z);
        },
        mul : function(m){
            return new Point3(this.x * m, this.y * m, this.z * m);
        },
        adds : function(v){
            this.x += v.x;
            this.y += v.y;
            this.z += v.z;
        },
        // ノルム
        norm : function(){
            return Math.sqrt(this.x * this.x + this.y * this.y + this.z * this.z);
        },
        // 正規化
        normalize : function(){
            var n = this.norm();
            if(n == 0) return new Point3(0,0,0);
            return new Point3(this.x / n, this.y / n, this.z / n);
        },
        // 外積
        cross : function(v){
            return new Point3(this.y * v.z - this.z * v.y,
                              this.z * v.x - this.x * v.z,
                              this.x * v.y - this.y * v.x);
        },
        // 内積
        dot : function(v){
            return this.x * v.x + this.y * v.y + this.z * v.z;
        },
        // ----
        // 透視投影。Pointを返す。（f : 焦点距離）
        to2d_pers : function(f){
            if(this.z >= f) return new Point();
            var d = f - this.z;
            return new Point(f * this.x / d,
                             f * this.y / d);
        },
        // 平行投影（z座標を0にした状態）Pointを返す
        to2d_ortho : function(){
            return this.toPoint();
        },
        // ----
        // z座標を除いた２次元座標の配列を返す（anchorの座標などに使うため）
        toArray : function(){
            return [this.x, this.y];
        },
        toPoint : function(){
            return new Point(this.x, this.y);
        },
        toString : function(){
            return "[" + this.x + "," + this.y + "," + this.z + "]";
        }
    }
    _vz = new Point3(0,0,1);

    // クォータニオンの掛け算の結果を返す
    function  qmul(q1, q2) {
        return [-q1[1]*q2[1] - q1[2]*q2[2] - q1[3]*q2[3] + q1[0]*q2[0],
                 q1[1]*q2[0] + q1[2]*q2[3] - q1[3]*q2[2] + q1[0]*q2[1],
                -q1[1]*q2[3] + q1[2]*q2[0] + q1[3]*q2[1] + q1[0]*q2[2],
                 q1[1]*q2[2] - q1[2]*q2[1] + q1[3]*q2[0] + q1[0]*q2[3]];
    }
    // クォータニオンのノルムを返す
    function qnorm(q){
        return Math.sqrt(q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3]);
    }
    // クォータニオンを正規化して返す
    function qnormalize(q){
        var n = qnorm(q);
        if(n == 0) return [1,0,0,0];
        return [q[0]/n, q[1]/n, q[2]/n, q[3]/n];
    }
    // 共役クォータニオンを返す
    function qconjugate(q){
        if(!q) return;
        return [q[0], -q[1], -q[2], -q[3]];
    }
    // n : 正規化された回転軸のベクトル(Point3)
    // t : 回転角度（ラジアン）
    function qget(n, t){
        t /= 2;
        var s = Math.sin(t);
        return [Math.cos(t), n.x*s, n.y*s, n.z*s];
    }
    // v : 回転したいベクトル(Point3)
    // q : クォータニオン
    // r : 共役クォータニオン。未指定の場合は生成される
    function qrot(v, q, r){
        if(!r) r = qconjugate(q);
        var p = [0, v.x, v.y, v.z];
        var qp = qmul(q, p);
        var qpr = qmul(qp, r);
        return new Point3(qpr[1], qpr[2], qpr[3]);
    }
    
    /* ------------------------------------
     * 汎用関数
     * ------------------------------------ */
    // 文字列sをfloatの値に変換する。変換できない場合は0を返す
    function tof(s){
        var n = parseFloat(s);
        return isNaN(n) ? 0 : n;
    }

    // 文字列sをintの値に変換する。変換できない場合は0を返す
    function toi(s){
        var n = parseInt(s);
        return isNaN(n) ? 0 : n;
    }

    // mmの値nをpostscript pointに換算して返す
    function mm2pt(n){
        return n * 2.83464567;
    }

    // 数値a,bが誤差_EPSILONを考慮して等しいと見なせればtrueを返す
    function isEqual(a,b){
        return Math.abs(a-b) < _EPSILON;
    }
    
    // 平面と、ベクトル(0,0,1) に平行な直線との交点のz座標を返す
    // n : Point3. 平面の法線
    // v : Point3. 平面が通る点
    // p : Point3. ベクトル(0,0,1) に平行な直線が通る点
    function getZ(n, v, p){
        if(!n || n.z == 0) return;
        return v.z - (n.x * (p.x - v.x) + n.y * (p.y - v.y)) / n.z;
    }

    function fixCoord(xy){
        xy[0] = Math.min(_MAX_COORD_VALUE, Math.max(-_MAX_COORD_VALUE, xy[0]));
        xy[1] = Math.min(_MAX_COORD_VALUE, Math.max(-_MAX_COORD_VALUE, xy[1]));
    }
    /* ------------------------------------
     * ItemInfo：回転の対象となるパスを保持する
     * ------------------------------------ */
    var ItemInfo = function(path, obj_idxs, obj_group){
        this.path = path;  // PathItem

        this.idxs = obj_idxs;

        this.normal;  // 面法線
        
        this.obj_group = obj_group;

        this.center;  // 点光源で使用

        // z座標補正
        this.z_max = 0;
        this.z_min = 0;
        this.z_max_index = 0;
        this.tmp_normal;  // 回転後の法線
        this.tmp_p;

        this.is_hidden;   // 視点方向から見て裏側を向いた面かどうか
        this.is_behind = false;  // 視点の後ろ側になって描画できない場合 true

        // 影
        this.shadows = [];  // PathItem[]
    }
    ItemInfo.prototype = {
        // 座標の初期化。
        init : function(){
            this._getNormal();
            this._getCenter();
        },
        // 面法線を求める
        _getNormal : function(){
            var a0 = _obj_vs[ this.idxs[0] ];
            var a1 = _obj_vs[ this.idxs[1] ];
            var a2 = _obj_vs[ this.idxs[2] ];

            this.normal = a1.sub(a0).cross(a2.sub(a1)).normalize();
        },
        // 中心（重心）を求める
        _getCenter : function(){
            var p = new Point3(0,0,0);
            for(var i = 0, iEnd = this.idxs.length; i < iEnd; i++){
                p.adds(_obj_vs[ this.idxs[i] ]);
            }
            this.center = p.mul( 1 / this.idxs.length);
        },
        // 回転を適用した頂点座標をパスに割り当てる
        applyRotation : function(){
            if(this.path.hidden){
                this.path.hidden = false;
            }
            this.is_behind = false;
            
            var pp = this.path.pathPoints;
            var z_max, z_min, z_max_index, tmp_p;
            
            for(var i = 0, iEnd = pp.length; i < iEnd; i++){
                var idx = this.idxs[i];
                var p = _obj_tmp_vs[idx];
                var z = p.z;
                if(i == 0){
                    z_max = z;
                    z_min = z;
                    z_max_index = idx;
                    tmp_p = p;
                } else {
                    if(z > z_max){
                        z_max = z;
                        z_max_index = idx;
                        tmp_p = p;
                    } else if(z < z_min){
                        z_min = z;
                    }
                }
                
                var r = _obj_view_vs[idx].toArray();
                if(r[0] == undefined){
                    this.is_behind = true;
                    this.path.hidden = true;
                    return;
                }
                fixCoord(r);
                pp[i].anchor = r;
                pp[i].rightDirection = r;
                pp[i].leftDirection = r;
            }
            this.z_max = z_max;
            this.z_min = z_min;
            this.z_max_index = z_max_index;
            this.tmp_p = tmp_p;
        },
    }

    /* ------------------------------------
     * rotateInit：mousedown時の処理
     * ------------------------------------ */
    function rotateInit(radius){
        if(_sources.length < 1) return;

        _radius = radius;
        
        for(var i = 0, iEnd = _sources.length; i < iEnd; i++){
            _sources[i].init();
        }
    }

    /* ------------------------------------
     * rotateMain：mousemove時の処理
     * ------------------------------------ */
    function rotateMain(x, y, radius){
        var isReset = _flags.mode == 1;
        var isFinish = _flags.mode == 2;
        
        // 共通して使うクォータニオンを取得
        var vlight;
        var default_light = false;
        if(!_light){
            default_light = true;
            // ライトの既定位置。なんとなく左上に
            setLight(-0.3 * _radius, -0.4 * _radius, _radius);
        }

        var q, r;
        if(isReset){
            _last_quaternion = undefined;
            _base_quaternion = undefined;
            vlight = _light.normalize();
        } else {
            if(isFinish){
                q = _last_quaternion || [1,0,0,0];
            } else {
                q = getRotMainQuaternion(x, y);
            }
            _last_quaternion = q;
            r = qconjugate(q);
            vlight = qrot(_light, q, r).normalize();
        }

        var plight;
        if(_opt.light_type == 1) plight = vlight.mul(_opt.light_distance);

        // 全ての頂点の回転後の位置を_obj_view_vsに入れる
        _obj_view_vs = [];
        _obj_tmp_vs = [];
        
        for(var vi = 0, viEnd = _obj_vs.length; vi < viEnd; vi++){
            var va = isReset
              ? _obj_vs[vi]
              : qrot(_obj_vs[vi], q, r);

            _obj_tmp_vs.push(va);

            _obj_view_vs.push(_opt.projection == 0
                              ? va.to2d_ortho()
                              : va.to2d_pers(_opt.focal_length));
            // pers は zが焦点距離を超える場合 (undefined, undefined)
        }

        var tint_range = _TINT_MAX - _opt.tint_min;

        for(var i = 0; i < _sources.length; i++){
            var item = _sources[i];

            // 面法線の回転と、面の明るさの決定
            var normal = isReset
              ? item.normal
                : qrot(item.normal, q, r).normalize();
            
            item.tmp_normal = normal;
            item.is_hidden = isHiddenSurface(_vz, normal);

            // 陰面は回転を適用せず隠す
            // → 影を描画するとき光源方向から見える面は必要なのでやっぱり回転する
            /*if(isFinish){
                item.path.hidden = false;
            } else {
                if(isHiddenSurface(vz, normal)){
                    item.path.hidden = true;
                    item.path.z_max = undefined;
                    continue;
                } else {
                    item.path.hidden = false;
                }
            }*/

            if(default_light){
                var tlight;
                if(_opt.light_type == 0){
                    tlight = normal.dot(vlight);
                } else {
                    var v = plight.sub(item.center).normalize();
                    tlight = v.dot(normal);
                }
                setTint(item.path, tlight, tint_range);
            }

            // パスに値を適用。z_max, z_minも決定
            item.applyRotation();
        }

        if(_sources.length > 1){
            fixZOrder(_opt.use_simple_z_sort_while_dragging);
        }
    }

    // rotateMain() で使うためのクォータニオンを返す
    function getRotMainQuaternion(x, y){
        // 呼び出し条件からd=0になることはない
        var d = Math.sqrt(x*x + y*y) / _radius;
        
        // 中心からの距離を sin の値として角度に換算する
        var t = -Math.asin(Math.min(d, 1));
        
        // 回転軸
        var n = new Point3(x, -y, 0).cross(_vz).normalize();
        
        var q = qget(n, t);
        
        if(_base_quaternion){
            q = qnormalize(qmul(q, _base_quaternion));
        }

        return q;
    }

    // 光源の位置を設定する
    function setLight(x, y, radius){
        d = Math.sqrt(x*x + y*y);
        if(d > radius){
            var rate = radius / d;
            x *= rate;
            y *= rate;
            d = 1;
        } else {
            d /= radius;
        }
        _light = new Point3(x, -y, radius * (1 - d));
    }

    // set light をチェックした状態でドラッグした時の処理
    function rotateLight(x, y, radius){
        if(x != undefined){
            setLight(x, y, radius);
            
            if(_base_quaternion){
                var r = qconjugate(_base_quaternion);
                _light = qrot(_light, r, _base_quaternion);
            }
        }

        vlight = _light.normalize();
        var plight;
        if(_opt.light_type == 1){
            plight = vlight.mul(_opt.light_distance);
        }
            
        var tint_range = _TINT_MAX - _opt.tint_min;

        if(_opt.light_type == 0){
            for(var i = 0; i < _sources.length; i++){
                var item = _sources[i];
                var tlight = item.normal.dot(vlight);
                setTint(item.path, tlight, tint_range);
            }
        } else {
            for(var i = 0; i < _sources.length; i++){
                var item = _sources[i];
                var v = plight.sub(item.center).normalize();
                var tlight = v.dot(item.normal);
                setTint(item.path, tlight, tint_range);
            }
        }
    }

    // 色（チント）を設定する
    // path : PathItem
    // tlight : cos値(面法線と光源ベクトルの内積)。-1～1。0のときもっとも明るい。
    // tint_range : _TINT_MAX - _opt.tint_min
    function setTint(path, tlight, tint_range){
        var tval;
        if(tlight >= 0){
            tval = (1 - tlight) * tint_range + _opt.tint_min;
        } else if(_opt.backlight_ratio > 0){
            tval = ((1 - _opt.backlight_ratio) + (_opt.backlight_ratio * (1 + tlight)))
              * tint_range + _opt.tint_min;
        } else {
            tval = 100;
        }
        path.fillColor.tint = Math.max(0, Math.min(100, tval));
    }

    /* ------------------------------------
     * rotateUpdate : mouseup時の処理
     * ------------------------------------ */
    function rotateUpdate(){
        //rotateMain();
        if(_last_quaternion){
            _base_quaternion = _last_quaternion;
        }
    }

    // 視点方向の逆を向いた面を判別するためのベクトルを返す
    function getHiddenSurfaceDetectVector(){
        if(_base_quaternion){
            return qrot(_vz, qconjugate(_base_quaternion), _base_quaternion);
        } else {
            return _vz;
        }
    }

    // vz : Point3
    // normal : Point3
    function isHiddenSurface(vz, normal){
        return vz.dot(normal) < 0;
    }

    // 視点方向の逆を向いた面と不可視パスを消去する
    function removeHiddenSurfaces(){
        if(!_obj_groups) return;
        
        var vz = getHiddenSurfaceDetectVector();

        var parent_grp = _sources[0].path.parent;

        if(_opt.save_hidden_surfaces){
            // 裏向きの面をグループに入れて残す
            var hidden_grp;
            for(var i = 0, iEnd = _sources.length; i < iEnd; i++){
                if(_sources[i].path.hidden){
                    // 不可視状態のパスは削除する
                    _sources[i].path.remove();
                } else if(_sources[i].is_hidden){
                    // 裏向きの面
                    if(!hidden_grp){
                        hidden_grp = activeDocument.activeLayer.groupItems.add();
                        hidden_grp.move(parent_grp, ElementPlacement.PLACEATEND);
                    }
                    _sources[i].path.move(hidden_grp, ElementPlacement.PLACEATBEGINNING);
                }
            }
        } else {
            // 裏向きの面を消去する
            for(var i = 0, iEnd = _sources.length; i < iEnd; i++){
                if(_sources[i].is_hidden){
                    // 裏向き
                    _sources[i].path.remove();
                } else if(_sources[i].path.hidden){
                    // 不可視状態
                    _sources[i].path.remove();
                }
            }
        }

        // パスがすべて削除されていたらグループを削除する
        if(parent_grp.pageItems.length < 1){
            parent_grp.remove();
        }
    }

    // z座標のみを比較する単純なソート方法
    // a, b : ItemInfo (.z_min, .z_max を使用）
    // 最小値を優先して比較
    function fixZOrder_SimpleSort_Min(a, b){
        return b.z_min - a.z_min || b.z_max - a.z_max;
    }
    // 最大値を優先して比較
    function fixZOrder_SimpleSort_Max(a, b){
        return b.z_max - a.z_max || b.z_min - a.z_min;
    }

    // 配列 r について、要素 val の最初のindexを返す。
    // val がない場合は -1 を返す。
    function arrayIndexOf(r, val){
        for(var i = 0, iend = r.length; i < iend; i++){
            if(r[i] == val) return i;
        }
        return -1;
    }

    // aとbの前後関係についての情報を返す
    // a, b : ItemInfo（.idxs, .z_min, .z_max, .z_max_index, .tmp_normal, .tmp_p を使用）
    function fixZOrder_Sort(a, b){
        if(a.z_max_index == b.z_max_index){
            if(isEqual(a.z_min, b.z_min)) return 0;
            if(a.z_min > b.z_min) return -1;
            return 1;
        } else if(isEqual(a.z_min, b.z_max) || a.z_min > b.z_max){
            return -1;
        } else if(isEqual(a.z_max, b.z_min) || a.z_max < b.z_min){
            return 1;
        } else {  // a.z_min < b.z_max, a.z_max > b.z_min
            if(arrayIndexOf(a.idxs, b.z_max_index) >= 0){
                return -1;
            } else if(arrayIndexOf(b.idxs, a.z_max_index) >= 0){
                return 1;
            }

            var state = surrounds(a, b);
            if(state == 0){
                var zs = hasIntersection(a, b, true);
                if(zs){
                    if(zs[0] == zs[1]){
                        return 0;
                    } else if(zs[0] == undefined || zs[1] == undefined){
                        return fixZOrder_SimpleSort_Min(a, b);
                    } else {
                        return zs[0] > zs[1] ? -1 : 1;
                    }
                } else {
                    return fixZOrder_SimpleSort_Min(a, b);
                }
            } else if(state == 1){  // a が b を囲んでいる
                var z = getZ(a.tmp_normal, a.tmp_p, b.tmp_p);
                if(z == undefined) return fixZOrder_SimpleSort_Min(a, b);
                return b.tmp_p.z - z;
            } else {  // state == -1, b が a を囲んでいる
                var z = getZ(b.tmp_normal, b.tmp_p, a.tmp_p);
                if(z == undefined) return fixZOrder_SimpleSort_Min(a, b);
                return z - a.tmp_p.z;
            }
        }
    }
    
    // 前後の並べ替え
    // 一応できてるようだけど何でできてるのかはよくわからない
    // orders : ItemInfo[]
    function sortOrdersMain(orders){
        orders.sort(function(a, b){ return fixZOrder_SimpleSort_Max(a,b) });
        if(_is_large_data){
            var progress = 10;
            showPercent(progress);
            var progress_range = 90 - progress;
        }

        var last_idx = orders.length - 1;
        for(var i = 1; i <= last_idx; i++){
            for(var j = last_idx; j >= i; j--){
                var src = orders[j];
                var dst = orders[j - 1];
                var d = fixZOrder_Sort(src, dst);
                if(d <= 0){
                    orders[j - 1] = src;
                    orders[j] = dst;
                }
            }
            if(_is_large_data) showPercent((i / orders.length) * progress_range + progress);
        }
    }

    // _sources のパス（視点方向を向いたもの）の前後関係を修正する
    function fixZOrder(use_simple_z_sort){
        if(_is_large_data) showPercent(0);
        var orders = [];
        var hiddens = [];
        for(var i = 0, iEnd = _sources.length; i < iEnd; i++){
            if(_sources[i].is_behind){
                // 不可視状態。処理対象外
            } else if(_sources[i].is_hidden){
                // 裏向き
                hiddens.push(_sources[i]);
            } else {
                orders.push(_sources[i]);
            }
        }
        
        if(_is_large_data) showPercent(5);
        
        var move_dst;
        if(orders.length > 1){
            if(use_simple_z_sort || _obj_groups.length < 2){
                orders.sort(function(a,b){ return fixZOrder_SimpleSort_Max(a,b) });
            } else {
                sortOrdersMain(orders);
            }
            
            if(_is_large_data) showPercent(90);

            // 先頭のものの後ろに順次移動する
            move_dst = orders[0].path;
            for(var i = 1, iEnd = orders.length; i < iEnd; i++){
                var item = orders[i];
                var move_src = item.path;
                move_src.move(move_dst, ElementPlacement.PLACEAFTER);
                var shadows = item.shadows;
                // 影がある場合は影も移動
                if(shadows.length > 0){
                    for(var j = 0, jend = shadows.length; j < jend; j++){
                        shadows[j].move(move_src, ElementPlacement.PLACEBEFORE);
                    }
                }
                move_dst = move_src;
            }
        }
        
        if(_is_large_data) showPercent(98);
        if(move_dst){
            for(var i = 0; i < hiddens.length; i++){
                hiddens[i].path.move(move_dst, ElementPlacement.PLACEAFTER);
            }
        }
        if(_is_large_data) showPercent(100);
    }

    // includefile objLoader_Shadow.jsxinc begin
    
    // 影に色を適用する
    // path : PatmItem
    function applyShadowColor(path){
        if(!_shadow_color){
            _shadow_color = new GrayColor();
            _shadow_color.gray = _opt.shadow_gray_value;
        }
        path.stroked = false;
        path.fillColor = _shadow_color;
        path.opacity = 50;
        path.blendingMode = BlendModes.MULTIPLY;
    }
    
    // 影
    var Shadow = function(path){  // path: PatmItem
        this.path = path;
        applyShadowColor(path);
    }
    Shadow.prototype = {
        // item : ItemInfo
        // q : クォータニオン
        // r : qの共役クォータニオン
        init : function(item, q, r){
            var v = _obj_tmp_vs[item.idxs[0]];
            var z = Math.abs(v.z);
            for(var i = 1, iEnd = item.idxs.length; i < iEnd; i++){
                var v1 = _obj_tmp_vs[item.idxs[i]];
                if(Math.abs(v1.z) > z){
                    v = v1;
                    z = v1.z;
                }
            }
    
            var pp = this.path.pathPoints;
            for(var i = 0, iEnd = pp.length; i < iEnd; i++){
                var p = new Point3().fromArray(pp[i].anchor);
                p.z = getZ(item.tmp_normal, v, p) || 0;
                p = qrot(p, q, r);
                var ar = (_opt.projection == 0
                  ? p.to2d_ortho()
                    : p.to2d_pers(_opt.focal_length)).toArray();
                fixCoord(ar);
                pp[i].anchor = ar;
                pp[i].rightDirection = ar;
                pp[i].leftDirection = ar;
            }
        }
    }
    
    // 面法線が光源方向を向いている面を tmp_items に入れる
    // q : クォータニオン
    // r : qの共役クォータニオン
    // tmp_items : ItemInfo[]. 処理対象をpushするための空の配列
    // tmp_hidden_items : ItemInfo. 裏面
    function getTmpItems(q, r, tmp_items, tmp_hidden_items){
        var tmp_idxs = [];   // 頂点index
        for(var i = 0, iEnd = _sources.length; i < iEnd; i++){
            var item = _sources[i];
            
            if(item.path.hidden) continue;
    
            // 面法線に逆quaternionを適用
            var n = qrot(item.normal, r, q);
            var ndot = n.normalize().dot(_vz);
            // こちらを向いている面のみ扱う
            if(ndot > 0){
                // 頂点indexをリストに追加
                tmp_idxs = tmp_idxs.concat(item.idxs);
                // ItemInfoをリストに追加
                item.tmp_normal = n;
                tmp_items.push(item);
            } else if(_opt.cast_shadow_on_hidden_faces
                      && item.is_hidden == false){
                // 視線方向からは見える面
                tmp_hidden_items.push(item);
            }
        }
        tmp_idxs = uniq(tmp_idxs);  // ex.0,1,10,11,2,4,5,6,7,8,9
        return tmp_idxs;
    }
    
    // _obj_vs の要素（頂点）に回転を適用し、
    // それぞれ _obj_tmp_vs の同じindexに保持する
    // tmp_idxs : int[]. 頂点のindex
    // q : クォータニオン
    // r : qの共役クォータニオン
    function getTmpVs(tmp_idxs, q, r){
        _obj_tmp_vs = new Array(_obj_vs.length);
        // 必要なindexのみ値を割り当てる
        for(var i = 0, iEnd = tmp_idxs.length; i < iEnd; i++){
            var idx = tmp_idxs[i];
            var p = qrot(_obj_vs[idx], r, q);
    
            // 光源の後ろ側になる点は除外
            if(_opt.projection == 1
               && p.z >= _opt.focal_length){
                p = undefined;
            }
            
            _obj_tmp_vs[idx] = p;
        }
    }
    
    // tmp_itemsから光源の後ろ側に頂点がある面を取り除く
    function verifyTmpItems(tmp_items){
        var verified_tmp_items = [];
        for(var i = 0, iend = tmp_items.length; i < iend; i++){
            var item = tmp_items[i];
            var ng = false;
            for(var vi = 0, viend = item.idxs.length; vi < viend; vi++){
                if(_obj_tmp_vs[item.idxs[vi]] == undefined){
                    ng = true;
                    break;
                }
            }
            if(!ng) verified_tmp_items.push(item);
        }
        return verified_tmp_items;
    }
    
    // 影を削除する
    function resetShadow(){
      if(_shadow_group_ground){
        _shadow_group_ground.remove();
        _shadow_group_ground = undefined;
      }
      
      for(var i = 0, iend = _sources.length; i < iend; i++){
          var item = _sources[i];
          for(var si = 0, siEnd = item.shadows.length; si < siEnd; si++){
              item.shadows[si].remove();
          }
          item.shadows = [];
      }
    }
    
    // 影関連の変数を削除する。（影の描画後、undoを行った際の処理）
    function clearShadows(){
        _shadow_group_ground = undefined;
        for(var i = 0, iend = _sources.length; i < iend; i++){
            _sources[i].shadows = [];
        }
    }
    
    // 影を作成する
    function castShadow(){
        showPercent(0);
        
        // 光源がz軸になるよう、全体を回転
        var t = Math.acos(_light.normalize().dot(_vz));
        var axis = _light.mul(-1).cross(_vz).normalize();
        var q = qget(axis, t);
        var r = qconjugate(q);
    
        var tmp_items, tmp_idxs;
    
        resetShadow();
        
        tmp_items = [];
        tmp_hidden_items = [];
        tmp_idxs = getTmpItems(q, r, tmp_items, tmp_hidden_items);
        getTmpVs(tmp_idxs, q, r);
        tmp_items = verifyTmpItems(tmp_items);
    
        // ItemInfoごとのz-orderを修正する
        tmp_items = fixTmpZOrder(tmp_items);
        showPercent(10);
        
        // 影を取得する
        var q1 = _base_quaternion
          ? qmul(_base_quaternion, q) : q;
        var r1 = qconjugate(q1);
    
        var progress = 10;
        var progress_range = 90 - progress;
    
        var saved_selection = app.activeDocument.selection;
        
        for(var i = 0, iEnd = tmp_items.length - 1; i < iEnd; i++){
            var src = tmp_items[i];
            
            for(var j = i + 1, jEnd = tmp_items.length; j < jEnd; j++){
                var dst = tmp_items[j];
    
                var state = surrounds(src, dst);
                
                if(state == -1){
                    var shadow_path = getRotatedPath(src);
                    shadow_path.move(dst.path, ElementPlacement.PLACEBEFORE);
                    castShadowSub(shadow_path, dst, q1, r1, dst);
                } else if(state == 1){
                    var shadow_path = getRotatedPath(dst);
                    castShadowSub(shadow_path, src, q1, r1, dst);
                } else if(hasIntersection(src, dst)){
                    // 共通部分の形状を取得する
                    var shadow_path = createShadowPath(src, dst);
                    if(shadow_path){
                        castShadowSub(shadow_path, dst, q1, r1, dst);
                    }
                }
            }
            showPercent((i / tmp_items.length) * progress_range + progress);
        }
    
        if(_opt.cast_shadow_on_hidden_faces){
            castShadowOnHiddenFaces(tmp_hidden_items);
        }
        showPercent(95);
        
        if(_base_quaternion){
            castShadowOnTheGround(r, q, r1, q1, tmp_items);
        }
        showPercent(100);
    
        app.activeDocument.selection = saved_selection;
    }
    
    function castShadowSub(shadow_path, dst1, q1, r1, dst2){
        var shadow = new Shadow(shadow_path);
        shadow.init(dst1, q1, r1);
        dst2.shadows.push(shadow.path);
    }
    
    // 光源方向と反対向きだが視線方向からは見える面に影をつける
    // (_opt.cast_shadow_on_hidden_faces == true の場合に実行）
    function castShadowOnHiddenFaces(tmp_hidden_items){
        for(var i = 0, iend = tmp_hidden_items.length; i < iend; i++){
            var src = tmp_hidden_items[i];
            var shadow_path = src.path.duplicate();
            shadow_path.move(src.path, ElementPlacement.PLACEBEFORE);
            applyShadowColor(shadow_path);
            src.shadows.push(shadow_path);
        }
    }
    
    // 地面（Z 座標をobjの下端とするXY平面）に落ちた影のパスを作る
    // r, q: 光源方向の共役クォータニオンとクォータニオン
    // r1, q1: q に物体の回転を加えたもの
    // tmp_items : ItemInfo[]. 処理対象
    function castShadowOnTheGround(r, q, r1, q1, tmp_items){
        // XY(XZ)平面への影
        var gnormal = qrot(new Point3(0,1,0), r, q);
        // 光源の位置が閾値より低い場合（影が長くなりすぎる・光源が地面の下になる）は描画しない
        var angle_threshold = Math.sin(_opt.cast_shadow_ground_angle_threshold * Math.PI / 180);
        if(gnormal.z < angle_threshold){
            alert(localize(resources.message.shadow_ground_angle_threshold));
            return;
        }
        
        var p0 = qrot(new Point3(0, _spec.vmin.y, 0), r, q);
        
        _shadow_group_ground = activeDocument.activeLayer.groupItems.add();
        for(var i = 0, iEnd = tmp_items.length; i < iEnd; i++){
            var src = tmp_items[i];
            // この時点ではアンカー数が同じパスが取得できればいい
            var shadow_path = src.path.duplicate();
            applyShadowColor(shadow_path);
            shadow_path.move(_shadow_group_ground, ElementPlacement.PLACEATEND);
            var pp = shadow_path.pathPoints;
            
            for(var j = 0, jEnd = pp.length; j < jEnd; j++){
                var idx = src.idxs[j];
                var p = _obj_tmp_vs[idx].clone();  // 逆回転させた座標
                p.z = getZ(gnormal, p0, p) || 0;  // 平面との交点のz座標を取得
                p = qrot(p, q1, r1);  // 逆回転を戻す＋視点へ回転
                
                var ar = (_opt.projection == 0
                          ? p.to2d_ortho()
                          : p.to2d_pers(_opt.focal_length)).toArray();
                if(ar[0] == undefined){  // 頂点が視点の後ろ側になるパスは描画しない
                    shadow_path.remove();
                    continue;
                }
                fixCoord(ar);
                pp[j].anchor = ar;
                pp[j].rightDirection = ar;
                pp[j].leftDirection = ar;
            }
        }
        _shadow_group_ground.move(tmp_items[0].path.parent, ElementPlacement.PLACEATEND);
    }
    
    // 配列 r の内容をソートしたうえで重複を取り除き、
    // 新しい配列を生成して返す
    // r : sort 可能で要素が等しいことを == で比較できる配列
    function uniq(r){
        if(r.length == 1) return r;
        r.sort();
        var result = [r[0]];
        var c = 0;
        for(var i = 1, iEnd = r.length; i < iEnd; i++){
            var ri = r[i];
            if(result[c] != ri){
                result.push(ri);
                c++;
            }
        }
        return result;
    }
    
    // tmp_itemsの要素を回転後の頂点のZ座標が大きい順に並べ替える
    // 実際のPathItemは並べ替えない
    // tmp_items: ItemInfo[]
    function fixTmpZOrder(tmp_items){
        if(tmp_items.length < 2) return tmp_items;
        var orders = [];
    
        for(var i = 0, iEnd = tmp_items.length; i < iEnd; i++){
            var item = tmp_items[i];
            var idx = item.idxs[0];
            var v = _obj_tmp_vs[idx];
    
            var min_z = v.z;
            var max_z = v.z;
            item.tmp_p = v;
            item.z_max_index = idx;
            
            for(var vi = 1, viEnd = item.idxs.length; vi < viEnd; vi++){
                idx = item.idxs[vi];
                v = _obj_tmp_vs[idx];
                if(v.z < min_z){
                    min_z = v.z;
                } else if(v.z > max_z){
                    max_z = v.z;
                    item.z_max_index = idx;
                    item.tmp_p = v;
                }
            }
            item.z_max = max_z;
            item.z_min = min_z;
            orders.push(item);
        }
    
        sortOrdersMain(orders);
        return orders;
    }
    
    // ItemInfo src, dst のパスの交差する部分を作成して返す
    // src, dst: ItemInfo
    function createShadowPath(src, dst){
        var shadow;
        //var saved_selection = app.activeDocument.selection;
        app.activeDocument.selection = [];
        
        var src_path = getRotatedPath(src);
        var dst_path = getRotatedPath(dst);
        
        var grp = activeDocument.activeLayer.groupItems.add();
        src_path.move(grp, ElementPlacement.PLACEATEND);
        dst_path.move(grp, ElementPlacement.PLACEATEND);
        grp.selected = true;
        app.executeMenuCommand("Live Pathfinder Crop");
        app.executeMenuCommand("expandStyle");
        
        grp = activeDocument.selection[0];
        if(grp.pageItems.length > 0){
            shadow = grp.pageItems[0];
            shadow.move(dst.path, ElementPlacement.PLACEBEFORE);
        }
        grp.remove();
        
        //app.activeDocument.selection = saved_selection;
        return shadow;
    }
    
    // パスを複製し、影の処理用に回転した頂点を適用したものを返す
    // src: ItemInfo
    function getRotatedPath(src){
        var path = src.path.duplicate();
        var pp = path.pathPoints;
        for(var i = 0, iEnd = pp.length; i < iEnd; i++){
            var r = _obj_tmp_vs[src.idxs[i]].toArray();
            fixCoord(r);
            pp[i].anchor = r;
            pp[i].rightDirection = r;
            pp[i].leftDirection = r;
        }
        return path;
    }
    
    // 凸包 points が p を囲んでいる場合 true を返す
    // 辺上にある場合も囲んでいるとする
    // points : Point3[] or Point[]
    // p: Point3 or Point
    function contains(points, p){
        var result = true;
        var n;
        for(var i = 0, iEnd = points.length - 1; i <= iEnd; i++){
            var j = i == iEnd ? 0 : i + 1;
            var p1 = points[i];
            var p2 = points[j];
            var d = (p1.x - p2.x) * (p.y - p1.y) + (p1.y - p2.y) * (p1.x - p.x);
            if(d == 0){  // 辺上
                result = true;
                break;
            }
            if(i == 0){
                n = d;
            } else {
                if((n < 0 && d > 0) || (n > 0 && d < 0)){
                    result = false;
                    break;
                }
            }
        }
        return result;
    }
    
    // item : ItemInfo
    // vs : Point3[]. 回転を適用した _obj_vs
    function convertPoints(item, vs){
        var points = [];
        for(var i = 0, iEnd = item.idxs.length; i < iEnd; i++){
            var idx = item.idxs[i];
            points.push(vs[idx]);
        }
        return points;
    }
    
    // src, dst について囲まれている状態についての情報を返す
    // src, dst : ItemInfo
    function surrounds(src, dst){
        // 1: src が dst を囲んでいる
        // -1: dst が src を囲んでいる
        // 0: どちらでもない
    
        // 回転後の座標値に変換。ここでやるのは非効率ではある
        var src_points = convertPoints(src, _obj_tmp_vs);
        var dst_points = convertPoints(dst, _obj_tmp_vs);
    
        var ng = false;
        for(var i = 0, iEnd = dst_points.length; i < iEnd; i++){
            if(contains(src_points, dst_points[i]) == false){
                ng = true;
                break;
            }
        }
        if(!ng) return 1;
        
        for(var i = 0, iEnd = src_points.length; i < iEnd; i++){
            if(contains(dst_points, src_points[i]) == false){
                ng = true;
                break;
            }
        }
        if(!ng) return -1;
    
        return 0;
    }
    
    // n が0以下か1以上の場合trueを返す
    function isOutOfRange01(n){
        if(n <= 0 || n >= 1) return true;
        // || isEqual(n, 0) || isEqual(n, 1)) return true; // 時間節約のためここまではやらない
        return false;
    }
    
    // src, dst のパスの辺の交点が見つかった場合 true を返す。
    // src, dst : ItemInfo
    // return_z_coordinates : 交点がある場合、[src側z座標値, dst側z座標値] を返す。
    //                        なければ false を返す。
    function hasIntersection(src, dst, return_z_coordinates){
        var result = false;
        var sidxs = src.idxs;
        var didxs = dst.idxs;
        
        for(var si = 0, siLast = sidxs.length - 1; si <= siLast; si++){
            var sj = si == siLast ? 0 : si + 1;
            var idx1 = sidxs[si];
            var idx2 = sidxs[sj];
            var p1 = _obj_tmp_vs[idx1];
            var p2 = _obj_tmp_vs[idx2];
    
            for(var di = 0, diLast = didxs.length - 1; di <= diLast; di++){
                var dj = di == diLast ? 0 : di + 1;
                var idx3 = didxs[di];
                if(idx1 == idx3 || idx2 == idx3) continue;
                var idx4 = didxs[dj];
                if(idx1 == idx4 || idx2 == idx4) continue;
                
                var p3 = _obj_tmp_vs[idx3];
                var p4 = _obj_tmp_vs[idx4];
    
                // Z軸方向から2Dとして見て線分の交点を求める
                // 参考 http://marupeke296.com/COL_2D_No10_SegmentAndSegment.html
                var v = p1.sub(p3).toPoint();
                var v1 = p4.sub(p3).toPoint();
                var v2 = p2.sub(p1).toPoint();
                
                var d = v1.cross(v2);
                if(d == 0) continue;
                
                var t2 = v.cross(v1) / d;
                if(isOutOfRange01(t2)) continue;
                
                var t1 = v.cross(v2) / d;
                if(isOutOfRange01(t1)) continue;
    
                if(return_z_coordinates){
                    // 交点(Point)
                    var p = p3.toPoint().add(v.add(v2.mul(t2)));
    
                    // 3Dでのz座標値を求める
                    var z1;  // src側のz座標
                    var z2;  // dst側のz座標
    
                    // src側
                    var v1 = p2.sub(p1);
                    if(v1.x != 0){
                        z1 = v1.z * (p.x - p1.x) / v1.x + p1.z;
                    } else if(v1.y != 0){
                        z1 = v1.z * (p.y - p1.y) / v1.y + p1.z;
                    }
                    // v1.x, v1.y が0 = v1がz軸に平行な場合
                    // いまのところ undefined を返している。考察不足
    
                    // dst側
                    var v2 = p4.sub(p3);
                    if(v2.x != 0){
                        z2 = v2.z * (p.x - p3.x) / v2.x + p3.z;
                    } else if(v2.y != 0){
                        z2 = v2.z * (p.y - p3.y) / v2.y + p3.z;
                    }
                    
                    result = [z1, z2];
                } else {
                    result = true;
                }
                break;
            }
            if(result) break;
        }
        return result;
    }
    // objLoader_Shadow.jsxinc end
    // includefile objLoader_UI.jsxinc begin
    /* ------------------------------------
     * showOptionDialog
     * ------------------------------------ */
    function showOptionDialog(){
        // dialogを表示
        var win = new Window("dialog", "ObjLoader");
        
        var projectionPanel = win.add("panel", undefined, localize(resources.option.projection)); //"投影法");
        projectionPanel.orientation = "row";
        projectionPanel.ortho = projectionPanel.add("radiobutton", undefined,
                                                            localize(resources.option.proj_pararell));  //"平行投影");
        projectionPanel.pers  = projectionPanel.add("radiobutton", undefined,
                                                            localize(resources.option.proj_perspective));  //"透視投影");
        projectionPanel.ortho.value = (_opt.projection == 0);
        projectionPanel.pers.value = (_opt.projection == 1);
    
        var textGrp = projectionPanel.add("group");
        textGrp.add("statictext", undefined, localize(resources.option.proj_focal_length));  //"焦点距離");
        var txtFocalLength = textGrp.add("edittext", undefined, _opt.focal_length);
        txtFocalLength.characters = 8;
    
        var chkGroup = win.add("group");
        chkGroup.orientation = "column";
        chkGroup.chkZOrderFix = chkGroup.add("checkbox", undefined, localize(resources.option.use_simple_z_sort_while_dragging));  //"ドラッグ中は簡易Z座標補正を行う");
        chkGroup.chkZOrderFix.value = _opt.use_simple_z_sort_while_dragging;
        chkGroup.chkSaveHidden = chkGroup.add("checkbox", undefined, localize(resources.option.save_hidden));  //"終了時に裏面を残す");
        chkGroup.chkSaveHidden.value = _opt.save_hidden_surfaces;
        
        // 倍率とサイズ(radiobuttonと入力欄）, scale_factor, size
        var scalePanel = win.add("panel", undefined, localize(resources.option.scale));  //"倍率(読込む前に設定)");
        scalePanel.alignChildren = "left";
        scalePanel.radioAutoGroup = scalePanel.add("group");
        scalePanel.radioAutoGroup.radio
          = scalePanel.radioAutoGroup.add("radiobutton", undefined, localize(resources.option.scale_auto));  //"自動：");
        scalePanel.radioAutoGroup.tbox
          = scalePanel.radioAutoGroup.add("edittext", undefined, _opt.size.toString());
        scalePanel.radioAutoGroup.tbox.characters = 8;
        scalePanel.radioAutoGroup.add("statictext", undefined, "mm");
        
        scalePanel.radioFixGroup = scalePanel.add("group");
        scalePanel.radioFixGroup.radio
          = scalePanel.radioFixGroup.add("radiobutton", undefined, localize(resources.option.scale_specified));  //"指定：");
        scalePanel.radioFixGroup.tbox
          = scalePanel.radioFixGroup.add("edittext", undefined, _opt.scale_factor.toString());
        scalePanel.radioFixGroup.tbox.characters = 8;
        scalePanel.radioFixGroup.add("statictext", undefined, localize(resources.option.scale_ratio));  //"倍");
    
        if(_opt.scale_type == 0){
            scalePanel.radioAutoGroup.radio.value = true;
            scalePanel.radioAutoGroup.tbox.enabled = true;
            scalePanel.radioFixGroup.radio.value = false;
            scalePanel.radioFixGroup.tbox.enabled = false;
        } else {
            scalePanel.radioAutoGroup.radio.value = false;
            scalePanel.radioAutoGroup.tbox.enabled = false;
            scalePanel.radioFixGroup.radio.value = true;
            scalePanel.radioFixGroup.tbox.enabled = true;
        }
        
        // tintの割合 tint_min
        var tintTextGroup = win.add("group");
        tintTextGroup.label = tintTextGroup.add("statictext", undefined, localize(resources.option.tint_min));  //"チント下限");
        tintTextGroup.tbox  = tintTextGroup.add("edittext", undefined, _opt.tint_min.toFixed(2));
        tintTextGroup.tbox.characters = 8;
        // 影の濃さ
        var shadowTextGroup = win.add("group");
        shadowTextGroup.label = shadowTextGroup.add("statictext", undefined, localize(resources.option.shadow_gray_value));  //"影の濃さ");
        shadowTextGroup.tbox  = shadowTextGroup.add("edittext", undefined, _opt.shadow_gray_value.toFixed(2));
        shadowTextGroup.tbox.characters = 8;
        // 大き目のデータと判断する面の数の閾値
        var largeDataThresholdGroup = win.add("group");
        largeDataThresholdGroup.label = largeDataThresholdGroup.add("statictext", undefined, localize(resources.option.large_data_threshold));  //"大き目のデータと判断する面数の閾値");
        largeDataThresholdGroup.tbox  = largeDataThresholdGroup.add("edittext", undefined, _opt.large_data_threshold);
        largeDataThresholdGroup.tbox.characters = 8;
    
        // 光源
        var lightPanel = win.add("panel", undefined, localize(resources.option.light));  //"光源");
        lightPanel.orientation = "column";
        lightPanel.group1 = lightPanel.add("group");
        lightPanel.group1.orientation = "row";
        lightPanel.group1.para = lightPanel.group1.add("radiobutton", undefined, localize(resources.option.light_pararell));  //"平行光源");
        lightPanel.group1.point  = lightPanel.group1.add("radiobutton", undefined, localize(resources.option.light_point));  //"点光源");
        lightPanel.group1.para.value = (_opt.light_type == 0);
        lightPanel.group1.point.value = (_opt.light_type == 1);
    
        lightPanel.group1.tboxGrp = lightPanel.add("group");
        lightPanel.group1.tboxGrp.add("statictext", undefined, localize(resources.option.light_dist));  //"点光源の距離");
        lightPanel.group1.tboxGrp.tbox
          = lightPanel.group1.tboxGrp.add("edittext", undefined, _opt.light_distance);
        lightPanel.group1.tboxGrp.tbox.characters = 8;
    
        lightPanel.group2 = lightPanel.add("group");
        lightPanel.group2.label = lightPanel.group2.add("statictext", undefined, localize(resources.option.backlight_ratio));  //"逆光の強さ(0.0-1.0)");
        lightPanel.group2.tbox = lightPanel.group2.add("edittext", undefined, _opt.backlight_ratio.toFixed(2));
        lightPanel.group2.tbox.characters = 8;
    
        lightPanel.add("statictext", undefined, localize(resources.option.light_point_notice));
        
        var btnGrp = win.add("group");
        var btn_ok = btnGrp.add("button", undefined, "OK");
        var btn_cancel = btnGrp.add("button", undefined, "Cancel");
    
        scalePanel.radioAutoGroup.radio.onClick = function(){
            scalePanel.radioFixGroup.radio.value = false;
            scalePanel.radioAutoGroup.tbox.enabled = true;
            scalePanel.radioFixGroup.tbox.enabled = false;
        }
        scalePanel.radioFixGroup.radio.onClick = function(){
            scalePanel.radioAutoGroup.radio.value = false;
            scalePanel.radioAutoGroup.tbox.enabled = false;
            scalePanel.radioFixGroup.tbox.enabled = true;
        }
    
        btn_ok.onClick = function(){
            // 検証
            var n = tof(txtFocalLength.text);
            if(isNaN(n) || n <= 0){
                alert(localize(resources.message_option.invalid_focal_length));
                return;
            }
            n = tof(scalePanel.radioAutoGroup.tbox.text);
            if(isNaN(n) || n <= 0){
                alert(localize(resources.message_option.invalid_scale_value_auto));
                return;
            }
            n = tof(scalePanel.radioFixGroup.tbox.text);
            if(isNaN(n) || n <= 0){
                alert(localize(resources.message_option.invalid_scale_value_specified));
                return;
            }
            n = tof(tintTextGroup.tbox.text);
            if(isNaN(n) || n < 0 || n > 100){
                alert(localize(resources.message_option.invalid_tint_min));
                return;
            }
            n = tof(shadowTextGroup.tbox.text);
            if(isNaN(n) || n < 0 || n > 100){
                alert(localize(resources.message_option.invalid_shadow_gray_value));
                return;
            }
            n = tof(largeDataThresholdGroup.tbox.text);
            if(isNaN(n) || n < 0){
                alert(localize(resources.message_option.invalid_large_data_threshold));
                return;
            }
            n = tof(lightPanel.group1.tboxGrp.tbox.text);
            if(isNaN(n)){
                alert(localize(resources.message_option.Invalid_point_light_distance));
                return;
            }
            n = tof(lightPanel.group2.tbox.text);
            if(isNaN(n)){
                alert(localize(resources.message_option.Invalid_backlight_ratio));
                return;
            }
            
            _opt.focal_length = tof(txtFocalLength.text);
            _opt.projection = projectionPanel.pers.value ? 1 : 0;
            _opt.use_simple_z_sort_while_dragging = chkGroup.chkZOrderFix.value;
            _opt.save_hidden_surfaces = chkGroup.chkSaveHidden.value;
            
            _opt.scale_type = scalePanel.radioAutoGroup.radio.value ? 0 : 1;
            _opt.size = tof(scalePanel.radioAutoGroup.tbox.text);
            _opt.scale_factor = tof(scalePanel.radioFixGroup.tbox.text);
    
            _opt.tint_min = tof(tintTextGroup.tbox.text);
            _opt.shadow_gray_value = tof(shadowTextGroup.tbox.text);
            _opt.large_data_threshold = tof(largeDataThresholdGroup.tbox.text);
            _opt.light_type = lightPanel.group1.para.value ? 0 : 1;
            _opt.light_distance = tof(lightPanel.group1.tboxGrp.tbox.text);
            _opt.backlight_ratio = tof(lightPanel.group2.tbox.text);
    
            win.close();
        }
            
        btn_cancel.onClick = function(){
            win.close();
        }
    
        win.show();
    }
    
    // アートボード中心の位置を取得する
    /* function findGlobalCenter(){
        var idx = activeDocument.artboards.getActiveArtboardIndex();
        var ab = activeDocument.artboards[idx];
        var rect = ab.artboardRect;  // left, top, right, bottom
        _center = new Point3((rect[0]+rect[2])/2, (rect[1]+rect[3])/2);
    } */
    
    /* ------------------------------------
     * メインウィンドウ
     * ------------------------------------ */
    // パーセント値を表示する。
    // win : showMainWindowで作成したWindow
    // value : パーセント値
    function showPercent(value){
        _win.status_text.text = value.toFixed(1) + "%";
        _win.update();
    }
    
    // メインウィンドウを表示する
    function showMainWindow(){
        if(documents.length < 1) return;
        if(activeDocument.activeLayer.locked){
            alert(localize(resources.message.unlock_active_layer));
            return;
        }
        
        _win = new Window("dialog", "OBJ loader");
        _win.status_text = _win.add("statictext", undefined, "---------------------");
        var panel = _win.add("panel", [0,0,_PANEL_SIZE,_PANEL_SIZE],undefined,{borderStyle:"none"});
        var g = panel.graphics;
        g.backgroundColor = g.newBrush(g.BrushType.SOLID_COLOR, [0.5,0.5,0.5]);
    
        var btnGroup = _win.add("group");
        var buttonOption = btnGroup.add("button", undefined, localize(resources.main_dialog.option));
        var buttonLoad = btnGroup.add("button", undefined, localize(resources.main_dialog.load_obj));
   
        // checkbox(光源設定)
        var chk_light = _win.add("checkbox", undefined, localize(resources.main_dialog.set_light));
    
        var btnGroup2 = _win.add("group");
        var buttonZOrderFix = btnGroup2.add("button", undefined, localize(resources.main_dialog.z_order_fix));
        buttonZOrderFix.enabled = false;  // OBJ読み込み後に有効にする
        // 閉じるときに裏面消去を行う。macでonCloseが有効か未確認。無効な場合はボタンで
        // var buttonClose = _win.add("button", undefined, "close");
        var buttonShadow = btnGroup2.add("button", undefined, localize(resources.main_dialog.cast_shadows));
        buttonShadow.enabled = false;  // OBJ読み込み後に有効にする
        
        var btnGroup3 = _win.add("group");
        var buttonClose = btnGroup3.add("button", undefined, localize(resources.main_dialog.close));
 
        var is_drawable = false;
        var is_in_process = false;
        var is_loaded = false;
    
        var is_button_shadow_first_time = true;
    
        var radius = _PANEL_SIZE / 2;
        // ドラッグ開始点
        var ox = radius;
        var oy = radius;
        // 最終描画時のポインタ位置
        var lastx, lasty;
    
        var movedEnough = function(cx, cy){
            // 最終ポインタ位置との差分
            var dx = cx - lastx;
            var dy = cy - lasty;
            return (dx*dx + dy*dy) > _MIN_DRAG_DIST;
        }
    
        var enableButtons = function(true_or_false){
            buttonShadow.enabled = true_or_false;
            buttonZOrderFix.enabled = true_or_false;
            buttonOption.enabled = true_or_false;
            buttonClose.enabled = true_or_false;
        }
        
        var enableControls = function(true_or_false){
            panel.enabled = true_or_false;
            chk_light.enabled = true_or_false;
            enableButtons(true_or_false);
        }
    
        var funcMousemove = function(e){
            if(is_drawable){
                var cx = e.clientX;
                var cy = e.clientY;
                
                if(movedEnough(cx, cy)){
                    try {
                        is_drawable = false;
                        is_in_process = true;
                        
                        lastx = cx;
                        lasty = cy;
                        _win.status_text.text = (cx - ox) + "," + (cy - oy);
                        var saved_txt;
                        if(_is_large_data) saved_txt = _win.status_text.text;
                        
                        if(chk_light.value){
                            if(_flags.drawn_light) app.undo();
                            _flags.mode = 0;  // 通常
                            
                            rotateLight(cx - ox, cy - oy, radius);
        
                            app.redraw();
                            _flags.drawn_light = true;
                        } else {
                            try{
                                if(_flags.drawn) app.undo();
                                _flags.mode = 0;  // 通常
                                
                                rotateMain(cx - ox, cy - oy, radius);
        
                                app.redraw();
                                if(_is_large_data){
                                    _win.status_text.text = saved_txt;
                                    _win.update();
                                }
                                _flags.drawn = true;
                            } catch(e){
                                alert(e);
                            }
                        }
                    } finally {
                        is_drawable = true;
                        is_in_process = false;
                    }
                }
            }
        };
    
        var funcMouseup = function(e){
            if(is_loaded){
                if(!_obj_groups) return;
                is_drawable = false;
                panel.removeEventListener("mousemove", funcMousemove);
    
                rotateUpdate();  // quaternionの更新のみ
                
                _flags.drawn = false;
                _flags.drawn_light = false;
                chk_light.value = false;
            }
        };
        
        var funcMousedown = function(e){
            if(is_loaded){
                if(!_obj_groups) return;
                ox = e.clientX;
                oy = e.clientY;
                lastx = ox;
                lasty = oy;
                panel.addEventListener("mousemove", funcMousemove);
                is_drawable = true;
            }
        };
        panel.addEventListener("mousedown", funcMousedown);
        panel.addEventListener("mouseup", funcMouseup);
        panel.addEventListener("mouseout", funcMouseup);
    
        
        // オプションボタン
        buttonOption.onClick = function(){
            if(is_in_process) return;
            
            var saved_light_type = _opt.light_type;
            var saved_light_distance = _opt.light_distance;
            var saved_backlight_ratio = _opt.backlight_ratio;
    
            showOptionDialog();
    
            buttonShadow.enabled = (is_loaded && _opt.light_type == 0);
            if(is_loaded){
                if(_opt.light_type != saved_light_type
                   || _opt.backlight_ratio != saved_backlight_ratio
                   || _opt.light_distance != saved_light_distance){
                    
                    rotateLight();
                    app.redraw();
                }
            }
            
            buttonZOrderFix.enabled = (is_loaded && _opt.use_simple_z_sort_while_dragging);
        }
    
        // Loadボタン
        buttonLoad.onClick = function(){
            is_loaded = false;
            try{
                this.enabled = false;
                enableControls(false);
                
                // リセット
                _obj_groups = [];
                _obj_groups = ObjRender();
                if(!_obj_groups || _obj_groups.length < 1){
                    _obj_groups = undefined;
                    return;
                }
                
                lastx = ox;
                lasty = oy;
                rotateInit(radius);
                _flags.mode = 1;  // reset
                rotateMain(0, 0, radius);
                app.redraw();
                
                // 読み込み完了したらボタンは押せないようにする
                is_loaded = true;
            } finally{
                if(!is_loaded) this.enabled = true;
                enableControls(true);
    
                buttonShadow.enabled = is_loaded;
                buttonZOrderFix.enabled = (is_loaded && _opt.use_simple_z_sort_while_dragging);
            }
        }
    
        // Closeボタン
        /* buttonClose.onClick = function(){
            //removeHiddenSurfaces();
            _win.close();
        } */
    
        buttonZOrderFix.onClick = function(){
            if(is_in_process) return;
    
            var saved_fix_state;
            try{
                saved_fix_state = _opt.use_simple_z_sort_while_dragging;
                enableControls(false);
                
                //if(_flags.drawn) app.undo();
                _flags.mode = 2;  // finish
                var saved_txt;
                if(_is_large_data) saved_txt = _win.status_text.text;
    
                _opt.use_simple_z_sort_while_dragging = false;
                rotateMain();
    
                if(_is_large_data){
                    _win.status_text.text = saved_txt;
                    _win.update();
                }
                app.redraw();
            } finally {
                _opt.use_simple_z_sort_while_dragging = saved_fix_state;
                enableControls(true);
            }
        }
    
        buttonShadow.onClick = function(){
            if(is_in_process) return;
            
            try{
                enableControls(false);
                
                if(this.text == localize(resources.main_dialog.cast_shadows)){
                    if(is_button_shadow_first_time){
                        alert(localize(resources.message.shadow_notice));
                        is_button_shadow_first_time = false;
                    }
                    if(_sources.length > _opt.large_data_threshold){
                        if(confirm(localize(resources.message.large_data_notice)) == false) return;
                    }
                    castShadow();
                    redraw();
                    this.text = localize(resources.main_dialog.undo_shadows);
                    // ドラッグできないようにする
                    is_loaded = false;
                } else {
                    app.undo();
                    clearShadows();
                    redraw();
                    this.text = localize(resources.main_dialog.cast_shadows);
                    // ドラッグできるようにする
                    is_loaded = true;
                }
            } catch(e) {
                alert(e.message);
            } finally {
               enableControls(true);
            }
        }
        
        buttonClose.onClick = function(){
            _win.close();
        }

        _win.onClose = function(){
            enableControls(false);
            
            removeHiddenSurfaces();
        }
        _win.show();
    }
    // objLoader_UI.jsxinc end
    // includefile objLoader_Resources.jsxinc begin
    /* ------------------------------------
     * Localize
     * ------------------------------------ */
    var resources = {
        message : {
            invalid_file_selection : {
                "en":"Select an OBJ file",
                "ja":"OBJ形式のファイルを選んでください" },
            fail_to_open_obj_to_load : {
                "en":"Failed to open the OBJ file to load.",
                "ja":"OBJファイルを読み取り用に開けませんでした。"},
            fail_to_open_mtl_to_load : {
                "en":"Failed to open the MTL file to load.\r",
                "ja":"MTLファイルを読み取り用に開けませんでした。\r"},
            error_on_loading_obj : {
                "en":"An error occured while reading the file.\r",
                "ja":"ファイルの読み取り中にエラーが発生しました。\r"},
            unlock_active_layer: {
                "en":"Please unlock active layer.",
                "ja":"アクティブレイヤーのロックを解除してください。"},
            shadow_notice : {
                "en":"Casting shadows. Rotation is locked while casting shadows.",
                "ja":"影を描画します。影を削除するまで回転はできません。"},
            large_data_notice : {
                "en":"It may take much time since it's a bit large data.",
                "ja":"データが大き目なので時間がかかるかもしれません。"},
            nothing_loaded : {
                "en":"nothing loaded", "ja":"何も読み込まれませんでした。"},
            warning_on_large_data_load : {
                "en":"About to load a large data.\rContinue loading?",
                "ja":"大き目のデータを読み込もうとしています。\r続行しますか？"},
            warning_on_large_data_load_plus : {
                "en":"(It will change z-order fix setting to 'While dragging'.)",
                "ja":"（ドラッグ中にZ座標簡易補正を行うよう設定を変更します。）"},
            shadow_ground_angle_threshold : {
                "en":"Shadows casted on the ground is not created because the location of the light source is too low. (_opt.cast_shadow_ground_angle_threshold)",
                "ja":"光源の位置が低すぎるため地面への影は描画されません。(_opt.cast_shadow_ground_angle_threshold)"},
        },
        message_option : {
            invalid_focal_length: {
                "en":"Invalid focal length value",
                "ja":"焦点距離の値が不適切です。"
            },
            invalid_scale_value_auto: {
                "en":"Invalid scale (auto) value",
                "ja":"倍率(自動/mm)の値が不適切です。"
            },
            invalid_scale_value_specified: {
                "en":"Invalid scale (specified) value",
                "ja":"倍率（指定/倍）の値が不適切です"
            },
            invalid_tint_min: {
                "en":"Invalid min. tint value.",
                "ja":"チント下限の値が不適切です"
            },
            invalid_shadow_gray_value: {
                "en":"Invalid shadow darkness value.",
                "ja":"影の濃さの値が不適切です"
            },
            invalid_large_data_threshold : {
                "en":"Invalid large data threshold.",
                "ja":"大きいデータと見なす閾値の値が不適切です"
            },
            Invalid_point_light_distance: {
                "en":"Invalid point light distance",
                "ja":"点光源の距離の値が不適切です"
            },
            Invalid_backlight_ratio : {
                "en":"Invalid backlighting strength",
                "ja":"光源の逆側からの光の強さが不適切です"
            }
        },
        option : {
            projection : {"en":"projection", "ja":"投影法"},
            proj_pararell : {"en":"pararell", "ja":"平行投影"},
            proj_perspective : {"en":"perspective", "ja":"透視投影"},
            proj_focal_length : {"en":"focal length", "ja":"焦点距離"},
            zorder : {"en":"simple z-order fix", "ja":"簡易的なZ座標補正"},
            zorder_none : {"en":"Never", "ja":"行わない"},
            zorder_drag : {"en":"While dragging", "ja":"ドラッグ中"},
            zorder_mouseup : {"en":"Even on mouseUp", "ja":"ドラッグ終了時も"},
            use_simple_z_sort_while_dragging : { "en":"simple z-order fix while dragging",
                "ja":"ドラッグ中は簡易的なZ座標補正を行う" },
            save_hidden : {"en":"save hidden surfaces on exit", "ja":"終了時に裏面を残す"},
            scale : {"en":"scale (set before load)", "ja":"倍率(読込む前に設定)"},
            scale_auto : {"en":"auto:", "ja":"自動："},
            scale_specified : {"en":"specified:", "ja":"指定："},
            scale_ratio : {"en":"times", "ja":"倍"},
            tint_min : {"en":"minimum tint value (0-100)", "ja":"チント下限(0-100)"},
            shadow_gray_value : {"en":"shadow darkness (0-100)", "ja":"影の濃さ(0-100)"},
            large_data_threshold : {"en":"large data threshold (faces count)", "ja":"大きいデータと見なす閾値（面数）"},
            light : {"en":"light", "ja":"光源"},
            light_pararell : {"en":"pararell", "ja":"平行光源"},
            light_point : {"en":"point", "ja":"点光源"},
            light_dist : {"en":"point light distance","ja":"点光源の距離"},
            backlight_ratio : {"en":"backlighting strength (0.0-1.0)", "ja":"光源の逆側からの光の強さ(0.0-1.0)"},
            light_point_notice : {"en":"'cast shadows' button is not available with point light",
                "ja":"点光源のとき影は作成できません。"},
        },
        main_dialog : {
            option : {"en":"option", "ja":"オプション"},
            load_obj : {"en":"load OBJ", "ja":"OBJ読み込み"},
            close : {"en":"close", "ja":"閉じる"},
            set_light : {"en":"set light", "ja":"光源位置変更"},
            z_order_fix : {"en":"z-order fix", "ja":"Z座標補正"},
            cast_shadows : {"en":"cast shadows", "ja":"影をつける"},
            undo_shadows : {"en":"undo shadows", "ja":"影を削除"},
        },
    };
    // objLoader_Resources.jsxinc end

    /* ------------------------------------
     * ObjLoader
     * ------------------------------------ */
    ObjLoader = {
        showMainWindow : function(){ showMainWindow(); },
    }
})();

ObjLoader.showMainWindow();
